import json
from .base import BaseResource


class StatusResource(BaseResource):
    def __init__(self, *args, **kwargs):
        super(StatusResource, self).__init__(*args, **kwargs)

    def on_get(self, req, resp):
        '''get status report for notebooks/jobs/reports or all'''
        type = req.params.get('type', '')
        if type == 'notebooks':
            resp.body = json.dumps(self.db.notebooks.status(req.context['user'], req.params, self.session))
        elif type == 'jobs':
            resp.body = json.dumps(self.db.jobs.status(req.context['user'], req.params, self.session))
        elif type == 'reports':
            resp.body = json.dumps(self.db.reports.status(req.context['user'], req.params, self.session))
        else:
            ret = {}
            ret['notebooks'] = self.db.notebooks.status(req.context['user'], req.params, self.session)
            ret['jobs'] = self.db.jobs.status(req.context['user'], req.params, self.session)
            ret['reports'] = self.db.reports.status(req.context['user'], req.params, self.session)
            resp.body = json.dumps(ret)



from .common import JobTask, JobCleanupTask, ReportTask  # noqa: F401
from .dokku import DokkuTask  # noqa: F401
from .nbconvert import NBConvertTask  # noqa: F401
from .papermill import PapermillTask  # noqa: F401
from .post import ReportPostTask  # noqa: F401
from .voila import VoilaTask  # noqa: F401



# CORS middleware
from .cors import CORS as CORSMiddleware  # noqa: F401

# Authentication middlewares
from .auth_required import AuthRequiredMiddleware  # noqa: F401
from .load_user import LoadUserMiddleware  # noqa: F401

# Form handling middleware
from .multipart import MultipartMiddleware  # noqa: F401

# No auth
from .none import NoUserMiddleware, NoAuthRequiredMiddleware  # noqa: F401

# SQL
from .sqla import SQLUserMiddleware, SQLAuthRequiredMiddleware, SQLAlchemySessionMiddleware  # noqa: F401



import falcon
import os
import os.path
import mimetypes
import jinja2
# from functools import lru_cache
from .base import BaseResource


# @lru_cache(20)
def read(file):
    '''read a file from local disk'''
    path = os.path.abspath(os.path.join(os.path.dirname(__file__), '..', 'assets', 'templates', file))
    if not os.path.exists(path):
        return None
    with open(path, 'r') as fp:
        return fp.read()


class HTMLResource(BaseResource):
    '''Falcon resource to service HTML files templated with jinja2'''
    def __init__(self, *args, **kwargs):
        super(HTMLResource, self).__init__(*args, **kwargs)

    def on_get(self, req, resp):
        '''Get templatized html and render'''
        if req.path == '' or req.path == '/':
            path = 'index.html'
        else:
            path = req.path

        filetype = mimetypes.guess_type(path, strict=True)[0]
        resp.content_type = filetype

        file = read(path)
        if file:
            tpl = jinja2.Template(file).render(user=req.context.get('user').name,
                                               baseurl=self.config.baseurl,
                                               apiurl=self.config.apiurl,
                                               loginurl=self.config.loginurl,
                                               logouturl=self.config.logouturl)
            resp.body = tpl
        else:
            resp.status = falcon.HTTP_404



import logging
from datetime import datetime
from paperboy.config import JobConfig
from paperboy.storage import JobStorage
from .base import BaseSQLStorageMixin, justid
from .models.user import UserSQL
from .models.notebook import NotebookSQL
from .models.job import JobSQL


class JobSQLStorage(BaseSQLStorageMixin, JobStorage):
    def status(self, user, params, session, *args, **kwargs):
        base = session.query(JobSQL) \
            .filter(JobSQL.userId == int(user.id))

        return {'total': base.count(),
                'production': base.filter(JobSQL.level == 'production').count(),
                'research': base.filter(JobSQL.level == 'research').count(),
                'development': base.filter(JobSQL.level == 'development').count(),
                'personal': base.filter(JobSQL.level == 'personal').count()}

    def form(self):
        '''Pass through to shared method in BaseSQLStorageMixin'''
        return self._form(JobConfig)

    def search(self, user, params, session, *args, **kwargs):
        '''Pass through to shared method in BaseSQLStorageMixin'''
        return self._search(JobSQL, 'Job', user, params, session, *args, **kwargs)

    def list(self, user, params, session, *args, **kwargs):
        '''Pass through to shared method in BaseSQLStorageMixin'''
        return self._list(JobSQL, 'jobs', user, params, session, *args, **kwargs)

    def detail(self, user, params, session, *args, **kwargs):
        '''Pass through to shared method in BaseSQLStorageMixin'''
        return self._detail(JobSQL, user, params, session, *args, **kwargs)

    def store(self, user, params, session, scheduler, *args, **kwargs):
        name = params.get('name')
        user_sql = session.query(UserSQL).get(int(user.id))
        notebookid = params.get('notebook')
        nb_sql = session.query(NotebookSQL).get(int(justid(notebookid)))
        notebook_config = nb_sql.to_config(self.config)

        start_time = datetime.strptime(params.get('starttime'), '%Y-%m-%dT%H:%M')
        interval = params.get('interval') or ''
        level = params.get('level') or ''
        created = datetime.now()
        modified = datetime.now()

        id = params.get('id')
        if id:
            id = justid(id)
            jb = session.query(JobSQL).filter(JobSQL.id == id).first()
            jb.name = name
            jb.start_time = start_time
            jb.interval = interval
            jb.level = level
            jb.modified = modified

        else:
            jb = JobSQL(name=name,
                        userId=user.id,
                        user=user_sql,
                        notebookId=notebookid,
                        notebook=nb_sql,
                        start_time=start_time,
                        interval=interval,
                        level=level,
                        created=created,
                        modified=modified)
        session.add(jb)

        # generate id
        session.flush()
        session.refresh(jb)

        logging.critical("Storing job {}".format(jb))
        job_config = jb.to_config(self.config)
        store = job_config.store()

        # autogenerate reports
        report_configs = self.db.reports.generate(user, params, session, job=job_config, *args, **kwargs)
        if report_configs:
            # reload to get accurate report count
            session.flush()
            session.refresh(jb)
            job_config = jb.to_config(self.config)
            store = job_config.store()

        scheduler.schedule(user, notebook_config, job_config, report_configs)
        return store

    def delete(self, user, params, session, scheduler, *args, **kwargs):
        # TODO only if allowed
        id = justid(params.get('id'))
        jb = session.query(JobSQL).filter(JobSQL.id == id).first()
        name = jb.name
        session.delete(jb)

        scheduler.unschedule(user, None, jb.to_config(self.config), [])

        return [{"name": "", "type": "p", "value": "Success!", "required": False, "readonly": False, "hidden": False},
                {"name": "", "type": "p", "value": "Successfully deleted job: " + name, "required": False, "readonly": False, "hidden": False}]



from airflow.models import BaseOperator
from airflow.utils.decorators import apply_defaults
from paperboy.utils import name_to_class


class DokkuOperator(BaseOperator):
    @apply_defaults
    def __init__(self, report, config, *args, **kwargs):
        super(DokkuOperator, self).__init__(*args, **kwargs)
        self.report = report
        self.nbconvert_task_id = self.task_id.replace('ReportPost', 'ReportNBConvert')

        self.config = name_to_class(config.get('config')).from_json(config)

    def execute(self, context):
        self.log.critical('report-post')

        task_instance = context['task_instance']
        output_nb = task_instance.xcom_pull(task_ids=self.nbconvert_task_id, key=self.task_id)
        self.log.critical(output_nb)

        outputter = self.config.clazz(self.config)
        outputter.write(self.report, output_nb, task_id=self.task_id)



import falcon
from six.moves.urllib_parse import urljoin


class SQLAlchemySessionMiddleware(object):
    '''variant of https://gitlab.com/skosh/falcon-helpers/blob/master/falcon_helpers/middlewares/sqla.py'''
    def __init__(self, sessionmaker=None):
        self.sessionmaker = sessionmaker

    def process_resource(self, req, resp, resource, params):
        '''initialize SQL Alchemy session and put into resource's `session` variable'''
        self.session = self.sessionmaker()
        resource.session = self.session

    def process_response(self, req, resp, resource, req_succeeded):
        '''If session is successful, commit, otherwise revert'''
        if not hasattr(resource, 'session'):
            return

        try:
            if not req_succeeded:
                self.session.rollback()
            else:
                self.session.commit()
        except Exception:
            self.session.remove()
            raise
        finally:
            self.session.close()


class SQLUserMiddleware(object):
    '''Middleware to fetch current user and put in resource's context'''
    def __init__(self, config, db, *args, **kwargs):
        self.config = config
        self.db = db

    def process_request(self, req, resp):
        '''Get user from auth token'''
        if req.context.get('auth_token') is None:
            # If the user doesnt have a token, redirect to login
            if req.path not in (urljoin(self.config.baseurl, self.config.loginurl), urljoin(self.config.baseurl, self.config.registerurl)) and \
               urljoin(self.config.baseurl, 'static') not in req.path:
                raise falcon.HTTPFound(urljoin(self.config.baseurl, self.config.loginurl))
        else:
            # try to get user

            # get sql session
            session = self.config.storage.sessionmaker()

            # query for user
            user = self.db.users.detail(req.context['auth_token'], req.params, session)

            if user is not None:
                # user not found
                req.context['user'] = user
                if req.path in (urljoin(self.config.baseurl, self.config.loginurl), urljoin(self.config.baseurl, self.config.registerurl)):
                    raise falcon.HTTPFound(self.config.baseurl)

            elif req.path not in (urljoin(self.config.baseurl, self.config.loginurl), urljoin(self.config.baseurl, self.config.registerurl), urljoin(self.config.baseurl, self.config.logouturl)) and \
                    urljoin(self.config.baseurl, 'static') not in req.path:
                raise falcon.HTTPFound(urljoin(self.config.baseurl, self.config.loginurl))


class SQLAuthRequiredMiddleware(object):
    '''Middleware to authenticate user with sqlalchemy'''
    def __init__(self,
                 config,
                 db):
        self.config = config
        self.db = db

    def process_request(self, req, resp):
        '''if auth token present, put into context'''
        token_value = req.cookies.get('auth_token', None)
        req.context['auth_token'] = token_value

    def process_resource(self, req, resp, resource, params):
        required = getattr(resource, 'auth_required', True)
        token_value = req.context.get('auth_token', None)
        token_value = None if isinstance(token_value, Exception) else token_value

        if required and not token_value:
            raise falcon.HTTPFound(urljoin(self.config.baseurl, self.config.loginurl))



'''Copied from https://gitlab.com/skosh/falcon-helpers due to conda-forge issues
Copyright (c) 2017 by Nicholas Zaccardi

Some rights reserved.

Redistribution and use in source and binary forms of the software as well
as documentation, with or without modification, are permitted provided
that the following conditions are met:

* Redistributions of source code must retain the above copyright
  notice, this list of conditions and the following disclaimer.

* Redistributions in binary form must reproduce the above
  copyright notice, this list of conditions and the following
  disclaimer in the documentation and/or other materials provided
  with the distribution.

* The names of the contributors may not be used to endorse or
  promote products derived from this software without specific
  prior written permission.

THIS SOFTWARE AND DOCUMENTATION IS PROVIDED BY THE COPYRIGHT HOLDERS AND
CONTRIBUTORS "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT
NOT LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR
A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT OWNER
OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL,
EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO,
PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR
PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF
LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING
NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS
SOFTWARE AND DOCUMENTATION, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH
DAMAGE.
'''
import falcon


def _default_failed(req, resp, **kwargs):
    raise falcon.HTTPFound('/auth/login')


class AuthRequiredMiddleware:
    """Requires a cookie be set with a valid JWT or fails

    Example:
        import falcon
        from falcon_helpers.middlewares.auth_required import AuthRequiredMiddleware

        class Resource:
            auth_required = True

            def on_get(self, req, resp):
                # ...

        def when_fails_auth(req, resp, token_value):
            raise TerribleException(token_value)

        api = falcon.API(
            middleware=[
                AuthRequiredMiddleware(when_fails=when_fails_auth)
            ]
        )

        api.add_route('/', Resource())

    Attributes:
        resource_param: The paramater to pull the boolean from

        context_key: the key the token will be found on the request

        when_fails: (function) A function to execute when the authentication
            fails
    """

    def __init__(self, resource_param='auth_required',
                 context_key='auth_token_contents',
                 when_fails=_default_failed):

        self.resource_param = resource_param
        self.context_key = context_key
        self.failed_action = when_fails

    def process_resource(self, req, resp, resource, params):
        required = getattr(resource, self.resource_param, True)
        token_value = req.context.get(self.context_key, None)

        token_value = None if isinstance(token_value, Exception) else token_value

        if required and not token_value:
            return self.failed_action(req=req, resp=resp, token_value=token_value)



import os
import os.path
import subprocess
from tempfile import NamedTemporaryFile
from venv import create


def create_virtualenv(requirements_txt, directory):
    '''Helper function to create a virtualenv from a requirements.txt file

    Args:
        requirements_txt (string): text of requirements.txt to use to build virtualenv
        directory (string): directory directory of created base virtualenv
    Returns:
        None
    '''
    create(directory, with_pip=True)

    source = os.path.join(directory, 'bin', 'activate')
    with NamedTemporaryFile(suffix='txt') as fp:
        fp.write(requirements_txt.encode('utf-8'))
        command = ['source', source, '&&', 'pip', 'install', '-r', fp.name]
        subprocess.call(command)



from sqlalchemy import Column, Integer, String, DateTime
from sqlalchemy import ForeignKey
from sqlalchemy.orm import relationship
from paperboy.config import JobConfig, JobMetadataConfig
from .base import Base


class JobSQL(Base):
    __tablename__ = 'jobs'
    id = Column(Integer, primary_key=True)
    name = Column(String)

    userId = Column(Integer, ForeignKey('users.id'))
    user = relationship('UserSQL', back_populates='jobs')

    notebookId = Column(Integer, ForeignKey('notebooks.id', ondelete='cascade'))
    notebook = relationship('NotebookSQL', back_populates='jobs')

    reports = relationship('ReportSQL', cascade='all,delete,delete-orphan', back_populates='job')

    start_time = Column(DateTime)
    interval = Column(String)
    level = Column(String, nullable=True)

    created = Column(DateTime)
    modified = Column(DateTime)

    @staticmethod
    def from_config(jb):
        # FIXME
        return JobSQL(name=jb.name,
                      userId=int(jb.user.id),
                      # user=jb.user,
                      notebookId=int(jb.notebook.id),
                      # notebook=jb.notebook,
                      reports=jb.meta.reports,
                      start_time=jb.meta.start_time,
                      interval=jb.meta.interval,
                      level=jb.meta.level,
                      created=jb.meta.created,
                      modified=jb.meta.modified)

    def to_config(self, config):
        ret = JobConfig(config)
        ret.id = 'Job-' + str(self.id)
        ret.name = self.name

        meta = JobMetadataConfig()
        meta.notebook = self.notebook.to_config(config)
        meta.username = self.user.name
        meta.userid = 'User-' + str(self.user.id)

        meta.start_time = self.start_time
        meta.interval = self.interval
        meta.level = self.level

        meta.reports = len(self.reports)

        meta.created = self.created
        meta.modified = self.modified

        ret.meta = meta
        return ret

    def __repr__(self):
        return "<Job(name='{}', user='{}', notebook='{}', start='{}', interval='{}'>".format(self.name, self.user, self.notebook, self.start_time, self.interval)



from .autocomplete import AutocompleteResource  # noqa: F401
from .config import ConfigResource  # noqa: F401
from .html import HTMLResource  # noqa: F401
from .job import JobResource, JobDetailResource  # noqa: F401
from .login import LoginResource  # noqa: F401
from .logout import LogoutResource  # noqa: F401
from .notebook import NotebookResource, NotebookDetailResource  # noqa: F401
from .register import RegisterResource  # noqa: F401
from .report import ReportResource, ReportDetailResource  # noqa: F401
from .static import StaticResource  # noqa: F401
from .scheduler import SchedulerResource  # noqa: F401
from .status import StatusResource  # noqa: F401



# -*- coding: utf-8 -*-
import os
import os.path
from ..base import BaseScheduler, interval_to_cron
from .schedule import LocalProcessScheduler


class LocalScheduler(BaseScheduler):
    def __init__(self, *args, **kwargs):
        '''Create a new local scheduler'''
        super(LocalScheduler, self).__init__(*args, **kwargs)
        self.scheduler = LocalProcessScheduler()

    def status(self, user, params, session, *args, **kwargs):
        '''Get status of job/report tasks'''
        type = params.get('type', '')

        # TODO
        gen = LocalScheduler.query()
        if type == 'jobs':
            return gen.get('jobs', [])
        elif type == 'reports':
            return gen.get('reports', [])
        else:
            return gen or {}

    @staticmethod
    def query():
        '''Get status of job/reports'''
        return {}

    def schedule(self, user, notebook, job, reports, *args, **kwargs):
        '''Schedule a DAG for `job` composed of `reports` to be run locally'''
        # create task
        working_dir = self.config.scheduler_config.working_directory
        job_dir = os.path.join(working_dir, str(job.id))
        os.makedirs(job_dir, exist_ok=True)
        self.scheduler.schedule(job, reports, job_dir, interval_to_cron(job.meta.interval, job.meta.start_time))

    def unschedule(self, user, notebook, job, reports, *args, **kwargs):
        '''Remove the DAG for `user` and `notebook` composed of `job` running `reports` from
            local'''
        if reports:
            # reschedule
            working_dir = self.config.scheduler_config.working_directory
            job_dir = os.path.join(working_dir, str(job.id))
            os.makedirs(job_dir, exist_ok=True)
            self.scheduler.schedule(job, reports, job_dir, interval_to_cron(job.meta.interval, job.meta.start_time))

        else:
            # delete
            self.scheduler.unschedule(job.id)



from abc import abstractmethod, ABCMeta
from random import choice
from six import with_metaclass
from ..config.base import Interval


def interval_to_cron(interval, start_time):
    if isinstance(interval, str):
        interval = Interval(interval)
    if interval == Interval.MINUTELY:
        # simple
        return '*/1 * * * *'
    elif interval == Interval.FIVE_MINUTES:
        return '{start_minute_mod_five}-59/5 * * * *'.format(start_minute_mod_five=start_time.minute % 5)
    elif interval == Interval.TEN_MINUTES:
        return '{start_minute_mod_ten}-59/10 * * * *'.format(start_minute_mod_ten=start_time.minute % 10)
    elif interval == Interval.THIRTY_MINUTES:
        return '{start_minute_mod_thirty}-59/30 * * * *'.format(start_minute_mod_thirty=start_time.minute % 30)
    elif interval == Interval.HOURLY:
        return '{start_minute} */1 * * *'.format(start_minute=start_time.minute)
    elif interval == Interval.TWO_HOURS:
        return '{start_minute} {start_hour_mod_two}-23/2 * * *'.format(start_minute=start_time.minute, start_hour_mod_two=start_time.hour % 2)
    elif interval == Interval.THREE_HOURS:
        return '{start_minute} {start_hour_mod_three}-23/3 * * *'.format(start_minute=start_time.minute, start_hour_mod_three=start_time.hour % 3)
    elif interval == Interval.SIX_HOURS:
        return '{start_minute} {start_hour_mod_six}-23/6 * * *'.format(start_minute=start_time.minute, start_hour_mod_six=start_time.hour % 6)
    elif interval == Interval.TWELVE_HOURS:
        return '{start_minute} {start_hour_mod_twelve}-23/12 * * *'.format(start_minute=start_time.minute, start_hour_mod_twelve=start_time.hour % 12)
    elif interval == Interval.DAILY:
        return '{start_minute} {start_hour} * * *'.format(start_minute=start_time.minute, start_hour=start_time.hour)
    elif interval == Interval.WEEKLY:
        return '{start_minute} {start_hour} * * {day_of_week}'.format(start_minute=start_time.minute, start_hour=start_time.hour, day_of_week=start_time.strftime('%a'))
    elif interval == Interval.MONTHLY:
        return '{start_minute} {start_hour} {start_day} * *'.format(start_minute=start_time.minute, start_hour=start_time.hour, start_day=start_time.day)
    else:
        raise Exception('Unknown interval: %s' % interval)


class BaseScheduler(with_metaclass(ABCMeta)):
    '''Scheduler abstract base class'''

    def __init__(self, config, db, *args, **kwargs):
        self.config = config
        self.db = db

    @abstractmethod
    def status(self, req, resp, *args, **kwargs):
        '''Get status for a given request
        Args:
            req (falcon Request)
            resp (falcon response)
        '''
        pass

    @abstractmethod
    def schedule(self, user, notebook, job, reports, *args, **kwargs):
        '''Schedule a job to generate reports for notebook and user
        Args:
            user (paperboy.config.User): user requesting the scheduling
            notebook (paperboy.config.Notebook): Notebook for the job
            job (paperboy.config.Job): Job parameters
            reports ([paperboy.config.Report]): Report configurations with parameters
        '''
        pass

    @abstractmethod
    def unschedule(self, user, notebook, job, reports, *args, **kwargs):
        '''Unschedule a job to generate reports for notebook and user
        Args:
            user (paperboy.config.User): user requesting the scheduling
            notebook (paperboy.config.Notebook): Notebook for the job
            job (paperboy.config.Job): Job parameters
            reports ([paperboy.config.Report]): Report configurations with parameters
        '''
        pass


class DummyScheduler(BaseScheduler):
    '''Dummy Scheduler class'''

    def __init__(self, config, db, *args, **kwargs):
        self.config = config
        self.db = db

    def status(self, user, params, session, *args, **kwargs):
        '''Get status for a given request
        Args:
            req (falcon Request)
            resp (falcon response)
        '''
        type = params.get('type', '')
        gen = DummyScheduler.fakequery()
        if type == 'jobs':
            return gen['jobs']
        elif type == 'reports':
            return gen['reports']
        else:
            return gen

    @staticmethod
    def fakequery():
        '''If airflow not present, fake the results for now so the UI looks ok'''
        ret = {'jobs': [], 'reports': []}
        for i in range(10):
            ret['jobs'].append(
                {'name': 'DAG-Job-{}'.format(i),
                 'id': 'Job-{}'.format(i),
                 'meta': {
                    'id':  'Job-{}'.format(i),
                    'execution': '01/02/2018 12:25:31',
                    'status': choice(['✔', '✘'])}
                 }
            )
            ret['reports'].append(
                {'name': 'Report-{}'.format(i),
                 'id': 'Report-{}'.format(i),
                 'meta': {
                    'run': '01/02/2018 12:25:31',
                    'status': choice(['✔', '✘']),
                    'type': choice(['Post', 'Papermill', 'NBConvert', 'Setup']),
                    }
                 }
            )
        return ret

    def schedule(self, user, notebook, job, reports, *args, **kwargs):
        '''Schedule a job to generate reports for notebook and user
        Args:
            user (paperboy.config.User): user requesting the scheduling
            notebook (paperboy.config.Notebook): Notebook for the job
            job (paperboy.config.Job): Job parameters
            reports ([paperboy.config.Report]): Report configurations with parameters
        '''
        pass

    def unschedule(self, user, notebook, job, reports, *args, **kwargs):
        '''Unschedule a job to generate reports for notebook and user
        Args:
            user (paperboy.config.User): user requesting the scheduling
            notebook (paperboy.config.Notebook): Notebook for the job
            job (paperboy.config.Job): Job parameters
            reports ([paperboy.config.Report]): Report configurations with parameters
        '''
        pass



from airflow.models import BaseOperator
from airflow.utils.decorators import apply_defaults


class NBConvertOperator(BaseOperator):
    @apply_defaults
    def __init__(self, report, *args, **kwargs):
        super(NBConvertOperator, self).__init__(*args, **kwargs)
        self.report = report
        self.papermill_task_id = self.task_id.replace('ReportNBConvert', 'ReportPapermill')
        self.report_post_task_iud = self.task_id.replace('ReportNBConvert', 'ReportPost')

    def execute(self, context):
        self.log.critical('nbconvert')

        task_instance = context['task_instance']
        papermilled = task_instance.xcom_pull(task_ids=self.papermill_task_id, key=self.task_id)

        if self.report['meta']['output'] != 'notebook':
            from paperboy.worker import run_nbconvert

            template = self.report['meta'].get('template', '')

            ret = run_nbconvert(self.report['meta']['notebook'],
                                papermilled,
                                self.report['meta']['output'],
                                template,
                                self.report['meta']['strip_code'],
                                )
        else:
            ret = papermilled

        task_instance.xcom_push(key=self.report_post_task_iud, value=ret)



from six import with_metaclass
from abc import abstractmethod, ABCMeta


class BaseOutput(with_metaclass(ABCMeta)):
    '''Abstract base class for Output types'''
    def __init__(self, config, *args, **kwargs):
        self.config = config

    @abstractmethod
    def write(self, report, *args, **kwargs):
        '''write report to output backend'''
        pass



import json
from .base import BaseResource


class AutocompleteResource(BaseResource):
    '''Autocompletion Falcon resource'''
    def __init__(self, *args, **kwargs):
        super(AutocompleteResource, self).__init__(*args, **kwargs)

    def on_get(self, req, resp):
        '''Get autocompletion.

        This method calls the .search() method on the storage managers users/notebooks/jobs/reports attribute
        '''
        resp.content_type = 'application/json'

        params = {}
        params['name'] = req.params.get('partial', None)
        params['count'] = 10
        user = req.context['user']

        type = req.params.get('type', None)

        if type is None:
            resp.body = json.dumps(self.db.notebooks.search(user, params, session=self.session) +
                                   self.db.jobs.search(user, params, session=self.session) +
                                   self.db.reports.search(user, params, session=self.session))

        elif type == 'notebooks':
            resp.body = json.dumps(self.db.notebooks.search(user, params, session=self.session))
        elif type == 'jobs':
            resp.body = json.dumps(self.db.jobs.search(user, params, session=self.session))
        elif type == 'reports':
            resp.body = json.dumps(self.db.reports.search(user, params, session=self.session))
        else:
            resp.body = 'No results for type {}'.format(type)



from .common import BaseTask
import json
import luigi
from luigi.parameter import Parameter, ParameterVisibility


class PapermillTask(BaseTask):
    report = Parameter(visibility=ParameterVisibility.HIDDEN)

    def __init__(self, *args, **kwargs):
        super(PapermillTask, self).__init__(*args, **kwargs)
        self._report = json.loads(self.report)

    def run(self):
        self.log.critical('papermill')

        from paperboy.worker import run_papermill
        ret = run_papermill(self._report['meta']['notebook'],
                            self._report['meta']['notebook_text'],
                            self._report['meta']['parameters'],
                            self._report['meta']['strip_code'])

        fp = self.output().open('w')
        fp.write(ret)
        fp.close()
        self._completed = True

    def output(self):
        return luigi.LocalTarget(self._report['name'])



"""Copied from https://github.com/yohanboniface/falcon-multipart due to conda-forge issues.

MIT License

Copyright (c) 2017 Yohan Boniface

Permission is hereby granted, free of charge, to any person obtaining a copy
of this software and associated documentation files (the "Software"), to deal
in the Software without restriction, including without limitation the rights
to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
copies of the Software, and to permit persons to whom the Software is
furnished to do so, subject to the following conditions:

The above copyright notice and this permission notice shall be included in all
copies or substantial portions of the Software.

THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
SOFTWARE.
"""
from io import BytesIO
import falcon
import cgi


class Parser(cgi.FieldStorage):
    pass


class MultipartMiddleware(object):

    def __init__(self, parser=None):
        self.parser = parser or Parser

    def parse(self, stream, environ):
        return self.parser(fp=stream, environ=environ)

    def parse_field(self, field):
        if isinstance(field, list):
            return [self.parse_field(subfield) for subfield in field]

        # When file name isn't ascii FieldStorage will not consider it.
        encoded = field.disposition_options.get('filename*')
        if encoded:
            # http://stackoverflow.com/a/93688
            encoding, filename = encoded.split("''")
            field.filename = filename
            # FieldStorage will decode the file content by itself when
            # file name is encoded, but we need to keep a consistent
            # API, so let's go back to bytes.
            # WARNING we assume file encoding will be same as filename
            # but there is no guaranty.
            field.file = BytesIO(field.file.read().encode(encoding))
        if getattr(field, 'filename', False):
            return field
        # This is not a file, thus get flat value (not
        # FieldStorage instance).
        return field.value

    def process_request(self, req, resp, **kwargs):

        if 'multipart/form-data' not in (req.content_type or ''):
            return

        # This must be done to avoid a bug in cgi.FieldStorage.
        req.env.setdefault('QUERY_STRING', '')

        # To avoid all stream consumption problem which occurs in falcon 1.0.0
        # or above.
        stream = (req.stream.stream if hasattr(req.stream, 'stream') else
                  req.stream)
        try:
            form = self.parse(stream=stream, environ=req.env)
        except ValueError as e:  # Invalid boundary?
            raise falcon.HTTPBadRequest('Error parsing file', str(e))

        for key in form:
            # TODO: put files in req.files instead when #418 get merged.
            req._params[key] = self.parse_field(form[key])



__all__ = ['config', 'middleware', 'output', 'resources', 'scheduler', 'server', 'storage', 'worker']

from . import config  # noqa: F401
from . import output  # noqa: F401
from . import utils  # noqa: F401
from ._version import VERSION as __version__  # noqa: F401



from airflow.models import BaseOperator
from airflow.utils.decorators import apply_defaults


class PapermillOperator(BaseOperator):
    @apply_defaults
    def __init__(self, report, *args, **kwargs):
        super(PapermillOperator, self).__init__(*args, **kwargs)
        self.report = report
        self.nbconvert_task_id = self.task_id.replace('ReportPapermill', 'ReportNBConvert')

    def execute(self, context):
        self.log.critical('papermill')

        from paperboy.worker import run_papermill
        ret = run_papermill(self.report['meta']['notebook'],
                            self.report['meta']['notebook_text'],
                            self.report['meta']['parameters'],
                            self.report['meta']['strip_code'])

        task_instance = context['task_instance']
        task_instance.xcom_push(key=self.nbconvert_task_id, value=ret)



from six.moves.urllib_parse import urljoin
from datetime import datetime
from traitlets import HasTraits, Unicode, Int, Instance, validate, TraitError
from .forms import Response, FormEntry, DOMEntry
from .base import Base, _INTERVAL_TYPES, _OUTPUT_TYPES, _REPORT_TYPES, _SERVICE_LEVELS
from .notebook import NotebookConfig


class JobMetadataConfig(HasTraits):
    '''Paperboy configuration object representing a Job (metadata component)'''
    notebook = Instance(NotebookConfig)
    username = Unicode()
    userid = Unicode()

    start_time = Instance(datetime)
    interval = Unicode()
    level = Unicode()

    @validate('interval')
    def _validate_interval(self, proposal):
        if proposal['value'] not in _INTERVAL_TYPES:
            raise TraitError('Unrecognized type : {}'.format(proposal['value']))
        return proposal['value']

    reports = Int()
    created = Instance(datetime)
    modified = Instance(datetime)

    def to_json(self, include_notebook=False):
        '''Convert JobMetadata to a JSON'''
        ret = {}
        ret['notebook'] = self.notebook.name
        if include_notebook:
            ret['notebook_text'] = self.notebook.meta.notebook
        # ret['notebookid'] = self.notebook.id
        ret['interval'] = self.interval
        ret['level'] = self.level
        ret['reports'] = self.reports
        ret['created'] = self.created.strftime('%m/%d/%Y %H:%M:%S')
        ret['modified'] = self.modified.strftime('%m/%d/%Y %H:%M:%S')
        return ret

    @staticmethod
    def from_json(jsn):
        '''Create JobMetadata from a JSON'''
        ret = JobMetadataConfig()
        for k, v in jsn.items():
            if k in ('created', 'modified'):
                ret.set_trait(k, datetime.strptime(v, '%m/%d/%Y %H:%M:%S'))
            else:
                ret.set_trait(k, v)
        return ret


class JobConfig(Base):
    '''Paperboy configuration object representing a Job'''
    name = Unicode()
    id = Unicode()
    meta = Instance(JobMetadataConfig)

    def to_json(self, include_notebook=False):
        '''Convert Job to a JSON'''
        ret = {}
        ret['name'] = self.name
        ret['id'] = self.id
        ret['meta'] = self.meta.to_json(include_notebook)
        return ret

    def form(self):
        '''Generate Form template for client from a Job object'''
        f = Response()
        f.entries = [
            FormEntry(name='name', type='text', label='Name', value=self.name, placeholder='Name for Job...', required=True),
            FormEntry(name='notebook', type='autocomplete', label='Notebook', url=urljoin(self.config.apiurl, 'autocomplete?type=notebooks&partial='), required=True),
            FormEntry(name='starttime', type='datetime', label='Start Time/Date', required=True),
            FormEntry(name='interval', type='select', label='Interval', options=_INTERVAL_TYPES, required=True),
            FormEntry(name='level', type='select', label='Level', options=_SERVICE_LEVELS, required=True),
            FormEntry(name='autogen', type='checkbox', label='Autogenerate reports', value='true', required=False),
            FormEntry(name='parameters_inline', type='textarea', label='Papermill params (.jsonl)', placeholder='Upload file or type here...', required=False),
            FormEntry(name='parameters', type='file', label='Papermill params (.jsonl)', required=False),
            FormEntry(name='options', type='label', label='Report options'),
            FormEntry(name='type', type='select', label='Type', options=_REPORT_TYPES, required=True),
            FormEntry(name='output', type='select', label='Output', options=_OUTPUT_TYPES, required=True),
            FormEntry(name='strip_code', type='select', label='Strip Code', options=['yes', 'no'], required=True),
            FormEntry(name='submit', type='submit', value='save', url=urljoin(self.config.apiurl, 'jobs?action=save'))
        ]
        return f.to_json()

    @staticmethod
    def from_json(jsn, config):
        '''Create Job from a JSON'''
        ret = JobConfig(config)
        ret.name = jsn['name']
        ret.id = jsn['id']
        ret.meta = JobMetadataConfig.from_json(jsn['meta'])
        ret.meta.notebook = NotebookConfig(config)  # FIXME
        return ret

    def edit(self):
        '''Generate Edit template for client from a Job object'''
        f = Response()
        f.entries = [
            FormEntry(name='name', type='text', value=self.name, label='Name', placeholder='Name for Job...', required=True),
            FormEntry(name='id', type='text', value=self.id, label='Id', hidden=True),
            FormEntry(name='notebook name', type='text', value=self.meta.notebook.name, label='Notebook', required=True, readonly=True),
            FormEntry(name='notebook', type='text', value=self.meta.notebook.id, required=True, readonly=True, hidden=True),
            FormEntry(name='starttime', value=self.meta.start_time.strftime('%Y-%m-%dT%H:%M'), type='datetime', label='Start Time/Date', required=True),
            FormEntry(name='interval', type='select', value=self.meta.interval, label='Interval', options=_INTERVAL_TYPES, required=True),
            FormEntry(name='level', type='select', value=self.meta.level, label='Level', options=_SERVICE_LEVELS, required=True),
            FormEntry(name='reports', type='text', value=str(self.meta.reports), readonly=True),
            FormEntry(name='save', type='submit', value='save', url=urljoin(self.config.apiurl, 'jobs?action=save')),
            FormEntry(name='delete', type='submit', value='delete', url=urljoin(self.config.apiurl, 'jobs?action=delete'))
        ]
        return f.to_json()

    def entry(self):
        '''Generate ListTable entry for client from a Job object'''
        f = Response()
        f.entries = [
            DOMEntry(name='name', type='label', value=self.name, label='Name'),
            DOMEntry(name='id', type='label', value=self.id, label='Id', hidden=True),
            DOMEntry(name='notebook', type='label', value=self.meta.notebook.name, label='Notebook'),
            DOMEntry(name='interval', type='label', value=self.meta.interval, label='Interval'),
            DOMEntry(name='level', type='label', value=self.meta.level, label='Level'),
            DOMEntry(name='reports', type='label', value=str(self.meta.reports), label='Reports'),
            DOMEntry(name='created', type='label', value=self.meta.created.strftime('%m/%d/%Y %H:%M:%S'), label='Created'),
            DOMEntry(name='modified', type='label', value=self.meta.modified.strftime('%m/%d/%Y %H:%M:%S'), label='Modified'),
            DOMEntry(name='delete', type='button', value='delete', label='delete', url=urljoin(self.config.apiurl, 'jobs?action=delete'))
        ]
        return f.to_json()

    def store(self):
        '''Generate response modal for client when saving a Job object'''
        ret = Response()
        ret.entries = [
            DOMEntry(type='h2', value='Success!'),
            DOMEntry(type='p', value='Successfully configured job: {}'.format(self.name)),
            DOMEntry(type='p', value='Notebook: {}'.format(self.meta.notebook.name)),
        ]
        if self.meta.reports:
            ret.entries.append(DOMEntry(type='p', value='Reports: {}'.format(self.meta.reports)))
        return ret.to_json()



from sqlalchemy import Column, Integer, String, Boolean, DateTime
from sqlalchemy import ForeignKey
from sqlalchemy.orm import relationship
from paperboy.config import ReportConfig, ReportMetadataConfig
from .base import Base


class ReportSQL(Base):
    __tablename__ = 'reports'
    id = Column(Integer, primary_key=True)
    name = Column(String)

    userId = Column(Integer, ForeignKey('users.id'))
    user = relationship('UserSQL', back_populates='reports')

    notebookId = Column(Integer, ForeignKey('notebooks.id', ondelete='cascade'))
    notebook = relationship('NotebookSQL', back_populates='reports')

    jobId = Column(Integer, ForeignKey('jobs.id', ondelete='cascade'))
    job = relationship('JobSQL', back_populates='reports')

    parameters = Column(String)
    type = Column(String)
    output = Column(String)
    strip_code = Column(Boolean)

    created = Column(DateTime)
    modified = Column(DateTime)

    @staticmethod
    def from_config(rp):
        # FIXME
        return ReportSQL(name=rp.name,
                         userId=int(rp.user.id),
                         # user=rp.user,
                         notebookId=int(rp.notebook.id),
                         # notebook=rp.notebook,
                         jobId=int(rp.job.id),
                         # job=rp.job,
                         parameters=rp.meta.parameters,
                         type=rp.meta.type,
                         output=rp.meta.output,
                         strip_code=rp.meta.strip_code,
                         created=rp.meta.created,
                         modified=rp.meta.modified)

    def to_config(self, config):
        ret = ReportConfig(config)
        ret.id = 'Report-' + str(self.id)
        ret.name = self.name

        meta = ReportMetadataConfig()
        meta.notebook = self.notebook.to_config(config)
        meta.job = self.job.to_config(config)
        meta.username = self.user.name
        meta.userid = 'User-' + str(self.user.id)

        meta.parameters = self.parameters
        meta.type = self.type
        meta.output = self.output
        meta.strip_code = self.strip_code

        # meta.template = self.template  # FIXME implement
        meta.created = self.created
        meta.modified = self.modified
        ret.meta = meta
        return ret

    def __repr__(self):
        return "<Report(name='%s')>" % (self.name)



from datetime import datetime
from sqlalchemy import create_engine
from sqlalchemy.orm import sessionmaker
from paperboy.storage.sqla.models.output import OutputSQL
from .base import BaseOutput


class SQLAOutput(BaseOutput):
    '''Output to sql database'''
    def __init__(self, config, *args, **kwargs):
        self.config = config
        self.engine = create_engine(self.storage.sql_url, echo=False)
        self.sessionmaker = sessionmaker(bind=self.engine)

    def write(self, report, output, *args, **kwargs):
        task_id = kwargs.get('task_id', '')
        name = task_id + '_' + datetime.now().strftime('%m-%d-%Y_%H-%M-%S')

        if report['meta']['output'] == 'notebook':
            name += '.ipynb'
        elif report['meta']['output'] == 'script':
            name += '.py'
        elif report['meta']['output'] == 'email':
            name += '.eml'
        elif report['meta']['output'] in ('pdf', 'html'):
            name += '.{}'.format(report['meta']['output'])

        with self.sessionmaker() as session:
            report_id = int(report['id'])
            created = datetime.now()

            out = OutputSQL(name=name,
                            reportId=report_id,
                            created=created)
            session.add(out)
            session.flush()



import json
from .base import BaseResource


class SchedulerResource(BaseResource):
    def __init__(self, *args, **kwargs):
        super(SchedulerResource, self).__init__(*args, **kwargs)

    def on_get(self, req, resp):
        '''Get scheduler status of job and reports'''
        resp.content_type = 'application/json'
        resp.body = json.dumps(self.scheduler.status(req.context['user'], req.params, self.session))



import logging


class BaseResource(object):
    '''Base falcon resource to handle shared attributes'''
    def __init__(self, config, db=None, scheduler=None):
        '''Base falcon resource
        Args:
            config (paperboy.config.application.Paperboy): Paperboy application config
            db (paperboy.storage.manager.StorageEngine): Interface to storage engine
            scheduler (paperboy.scheduler.base.BaseScheduler): Scheduler
        '''
        self.config = config
        self.db = db
        self.scheduler = scheduler
        self.logger = logging.getLogger('paperboy.' + __name__)
        self.session = None  # May be overridden by middleware



from airflow.models import BaseOperator
from airflow.utils.decorators import apply_defaults


class JobOperator(BaseOperator):
    @apply_defaults
    def __init__(self, job, *args, **kwargs):
        super(JobOperator, self).__init__(*args, **kwargs)
        self.job = job

    def execute(self, context):
        self.log.critical('job')


class JobCleanupOperator(BaseOperator):
    @apply_defaults
    def __init__(self, job, *args, **kwargs):
        super(JobCleanupOperator, self).__init__(*args, **kwargs)
        self.job = job

    def execute(self, context):
        self.log.critical('job-cleanup')


class ReportOperator(BaseOperator):
    @apply_defaults
    def __init__(self, report, *args, **kwargs):
        super(ReportOperator, self).__init__(*args, **kwargs)
        self.report = report

    def execute(self, context):
        self.log.critical('report')



import falcon
import jinja2
from .base import BaseResource
from .html import read


class LoginResource(BaseResource):
    '''Falcon resource for user authentication'''

    auth_required = False

    def __init__(self, *args, **kwargs):
        super(LoginResource, self).__init__(*args, **kwargs)

    def on_get(self, req, resp):
        '''Get login page'''
        resp.content_type = 'text/html'
        file = read('login.html')
        tpl = jinja2.Template(file).render(baseurl=self.config.baseurl,
                                           apiurl=self.config.apiurl,
                                           loginurl=self.config.loginurl,
                                           include_register=self.config.include_register,
                                           registerurl=self.config.registerurl,
                                           include_password=self.config.include_password)
        resp.body = tpl

    def on_post(self, req, resp):
        '''Log user in using authentication backend'''
        token = self.db.users.login(None, req.params, self.session)
        user = self.db.users.detail(token, req.params, self.session)

        if token and user:
            # setup token and set auth cookie
            req.context['auth_token'] = token
            req.context['user'] = user
            resp.set_cookie('auth_token', token, max_age=self.config.token_timeout, path='/', secure=not self.config.http)
            resp.status = falcon.HTTP_302
            resp.set_header('Location', self.config.baseurl)

        else:
            # rerender login page
            resp.content_type = 'text/html'
            file = read('login.html')
            tpl = jinja2.Template(file).render(baseurl=self.config.baseurl,
                                               apiurl=self.config.apiurl,
                                               loginurl=self.config.loginurl,
                                               include_register=self.config.include_register,
                                               registerurl=self.config.registerurl,
                                               include_password=self.config.include_password)
            resp.body = tpl



from traitlets import HasTraits, Unicode, Bool
from ..storage.sqla import UserSQLStorage, NotebookSQLStorage, JobSQLStorage, ReportSQLStorage


class StorageConfig(HasTraits):
    '''Base config for storage backend'''
    type = Unicode()
    user_storage = None
    notebook_storage = None
    job_storage = None
    report_storage = None


class SQLAStorageConfig(StorageConfig):
    '''Config for SQL Alchemy storage'''
    type = 'SQLA'
    sql_url = Unicode(default_value='sqlite:///paperboy.db', help="SQL Alchemy url").tag(config=True)
    engine = None
    sessionmaker = None
    sql_user = Bool(default_value=True)

    user_storage = UserSQLStorage
    notebook_storage = NotebookSQLStorage
    job_storage = JobSQLStorage
    report_storage = ReportSQLStorage



import gunicorn.app.base
from gunicorn.six import iteritems


class FalconGunicorn(gunicorn.app.base.BaseApplication):
    '''Utility to deploy falcon.API on gunicorn'''

    def __init__(self, app, options=None):
        '''Constructor
        Args:
            app (falcon.API): a routed falcon API
            options (dict): a set of options for gunicorn (e.g. workers, port, etc)
        '''
        self.options = options or {}
        self.application = app
        super(FalconGunicorn, self).__init__()

    def load_config(self):
        config = dict([(key, value) for key, value in iteritems(self.options)
                       if key in self.cfg.settings and value is not None])
        for key, value in iteritems(config):
            self.cfg.set(key.lower(), value)

    def load(self):
        return self.application



import logging
from luigi import Task
from luigi.parameter import Parameter, DateParameter, ParameterVisibility


class BaseTask(Task):
    task_id = Parameter()

    def __init__(self, *args, **kwargs):
        super(BaseTask, self).__init__(*args, **kwargs)
        self._reqs = []
        self.log = logging  # FIXME
        self._completed = False

    def requires(self):
        return self._reqs

    def complete(self):
        if self._reqs:
            if isinstance(self._reqs, list):
                return all(r.complete() for r in self._reqs) and self._completed
            return self._reqs.complete() and self._completed
        return self._completed


class JobTask(BaseTask):
    job = Parameter(visibility=ParameterVisibility.HIDDEN)

    def __init__(self, *args, **kwargs):
        super(JobTask, self).__init__(*args, **kwargs)

    def run(self):
        self.log.critical('job')
        self._completed = True


class JobCleanupTask(BaseTask):
    job = Parameter(visibility=ParameterVisibility.HIDDEN)
    interval = Parameter()
    start_date = DateParameter()
    owner = Parameter()
    email = Parameter()
    time = DateParameter()

    def __init__(self, *args, **kwargs):
        super(JobCleanupTask, self).__init__(*args, **kwargs)

    def run(self):
        self.log.critical('job-cleanup')
        self._completed = True


class ReportTask(BaseTask):
    report = Parameter(visibility=ParameterVisibility.HIDDEN)

    def __init__(self, *args, **kwargs):
        super(ReportTask, self).__init__(*args, **kwargs)

    def run(self):
        self.log.critical('report')
        self._completed = True



import requests
from .base import BaseScheduler


class RemoteScheduler(BaseScheduler):
    '''Proxy methods to a remote worker instance'''
    def __init__(self, *args, **kwargs):
        super(RemoteScheduler, self).__init__(*args, **kwargs)

    def status(self, user, params, session, *args, **kwargs):
        # FIXME async/celery
        return requests.get(self.config.scheduler.status_url, params=params).json()

    def schedule(self, user, notebook, job, reports, *args, **kwargs):
        # FIXME async/celery
        params = {'user': user.to_json(), 'notebook': notebook.to_json(), 'job': job.to_json(), 'reports': [r.to_json() for r in reports]}
        return requests.post(self.config.scheduler.schedule_url, params=params).json()

    def unschedule(self, user, notebook, job, reports, *args, **kwargs):
        return self.schedule(user, notebook, job, reports, *args, **kwargs)



from .models.base import Base  # noqa: F401
from .job import JobSQLStorage  # noqa: F401
from .notebook import NotebookSQLStorage  # noqa: F401
from .report import ReportSQLStorage  # noqa: F401
from .user import UserSQLStorage  # noqa: F401



import base64
import emails
from bs4 import BeautifulSoup
from datetime import datetime
from jupyterlab_email.attachments import CUSTOM_TAG
from six import iteritems
from .base import BaseOutput


def make_email(name,
               data,
               from_,
               type='email',
               subject='',
               header='',
               footer=''):
    '''Helper method to convert jupyter notebook into an email

    Args:
        path (str): path to notebook
        model (str): notebook itself (in case deployment strips outputs or
                     notebook not available except through ContentsManager)
        from_ (str): address to send the email from
        type (str): type to convert notebook to
        template (str): template to use when converting notebook
        code (boolean): include input cells in notebook
        subject (str): subject of email
        header (str): html header to inject
        footer (str): html footer to inject
        also_attach (str): also attach pdf/html/both
        postprocessor (function): run postprocessor on soup
    Returns:
        emails.html: HTML Email to send
    '''
    if type == 'email':
        type_to = 'html'
    elif type == 'html attachment':
        type_to = 'html'
    elif type == 'pdf attachment':
        type_to = 'pdf'
    else:
        raise Exception('Type not recognized')

    if type == 'email':
        soup = BeautifulSoup(data, 'html.parser')

        # strip markdown links
        for item in soup.findAll('a', {'class': 'anchor-link'}):
            item.decompose()

        # strip matplotlib base outs
        for item in soup.find_all('div', class_='output_text output_subarea output_execute_result'):
            for c in item.contents:
                if '&lt;matplotlib' in str(c):
                    item.decompose()

        # remove dataframe table borders
        for item in soup.findAll('table', {'border': 1}):
            item['border'] = 0
            item['cellspacing'] = 0
            item['cellpadding'] = 0

        # extract imgs for outlook
        imgs = soup.find_all('img')
        imgs_to_attach = {}

        # attach main part
        for i, img in enumerate(imgs):
            if not img.get('localdata'):
                continue
            imgs_to_attach[img.get('cell_id') + '_' + str(i) + '.png'] = base64.b64decode(img.get('localdata'))
            img['src'] = 'cid:' + img.get('cell_id') + '_' + str(i) + '.png'
            # encoders.encode_base64(part)
            del img['localdata']

        attaches = soup.find_all(CUSTOM_TAG)
        att_to_attach = {}

        # attach custom attachments
        for i, att in enumerate(attaches):
            if not att.get('localdata'):
                continue
            filename = att.get('filename')
            if filename.endswith('.png') or \
               filename.endswith('.xls') or \
               filename.endswith('.xlsx') or \
               filename.endswith('.pdf'):
                att_to_attach[filename] = base64.b64decode(att.get('localdata'))
            else:
                att_to_attach[filename] = att.get('localdata')
            att.decompose()

        # attach header/footer
        if header or footer:
            head = soup.find('div', {'class': 'header'})
            foot = soup.find('div', {'class': 'footer'})
            head.append(BeautifulSoup(header, 'html.parser'))
            foot.append(BeautifulSoup(footer, 'html.parser'))

        # assemble email soup
        soup = str(soup)
        message = emails.html(charset='utf-8', subject=subject, html=soup, mail_from=from_)

        for img, data in iteritems(imgs_to_attach):
            message.attach(filename=img, content_disposition="inline", data=data)

        for att, data in iteritems(att_to_attach):
            message.attach(filename=att, content_disposition="inline", data=data)

        return message

    message = emails.html(subject=subject, html='<html>Attachment: %s.%s</html>' % (name, type_to), mail_from=from_)
    message.attach(filename=name + '.' + type_to, data=data)
    return message


def email(message, to, username, password, domain, host, port):
    '''Email a given message using smtp server details
    Args:
        to (str): who to send notebook to
        username (str): email account username
        password (str): email account password
        domain (str): email account provider
        host (str): smtp host
        port (str): smtp port
    '''
    r = message.send(to=to,
                     smtp={'host': host,
                           'port': port,
                           'ssl': True,
                           'user': username,
                           'password': password})
    if r.status_code != 250:
        print(r)
        raise Exception('Email exception! Check username and password')
    return r


class EmailOutput(BaseOutput):
    '''Email output type'''
    def __init__(self, config, *args, **kwargs):
        self.config = config

    def write(self, report, output, *args, **kwargs):
        '''write a given notebook as an email'''
        task_id = kwargs.get('task_id', '')

        name = task_id + '_' + datetime.now().strftime('%m-%d-%Y_%H-%M-%S')
        msg = make_email(name,
                         output,
                         from_=self.config.from_,
                         type=report['meta']['output'],
                         subject=name,
                         header=self.config.header,
                         footer=self.config.footer)
        email(msg, self.config.to, self.config.username, self.config.password, self.config.domain, self.config.host, self.config.port)



from sqlalchemy import Column, Integer, String, DateTime, LargeBinary
from sqlalchemy import ForeignKey
from sqlalchemy.orm import relationship
from .base import Base


class OutputSQL(Base):
    __tablename__ = 'outputs'
    id = Column(Integer, primary_key=True)
    name = Column(String)

    reportId = Column(Integer, ForeignKey('reports.id'))
    report = relationship('ReportSQL', back_populates='outputs')

    created = Column(DateTime)
    data = LargeBinary()

    def __repr__(self):
        return "<Output(name='%s')>" % (self.name)



import os
import os.path
from traitlets import HasTraits, Unicode
from ..output import LocalOutput
from ..utils import class_to_name, name_to_class


class OutputConfig(HasTraits):
    '''Base class for configuring output'''
    type = Unicode()
    pass


class LocalOutputConfig(OutputConfig):
    '''Output report to local filesystem'''
    type = 'local'
    dir = Unicode(default_value=os.path.expanduser('~/Downloads'))
    clazz = LocalOutput

    def to_json(self):
        ret = {}
        ret['type'] = self.type
        ret['dir'] = self.dir
        ret['clazz'] = class_to_name(self.clazz)
        ret['config'] = class_to_name(self.__class__)
        return ret

    @staticmethod
    def from_json(jsn):
        ret = LocalOutputConfig()
        ret.type = jsn.get('type')
        ret.dir = jsn.get('dir', '.')
        ret.clazz = name_to_class(jsn.get('clazz'))
        return ret



import json
import os
import os.path
import nbformat
from six import string_types
from papermill import execute_notebook
try:
    from tempfile import TemporaryDirectory
except ImportError:
    from backports.tempfile import TemporaryDirectory


def run(nb_name, nb_text, parameters, hide_input):
    '''Run the notebook and return the text

    Args:
        nb_name (string): Name of notebook
        nb_text (string): nbformat json of text of notebook to convert
        paramters (string): json parameters to use for papermill
        hide_input (boolean): hide code
    '''
    # validation for papermill
    nb = nbformat.reads(nb_text, 4)

    with TemporaryDirectory() as tempdir:
        in_file = os.path.join(tempdir, '{}.ipynb'.format(nb_name))
        out_file = os.path.join(tempdir, '{}_out.ipynb'.format(nb_name))

        nbformat.write(nb, in_file)

        if isinstance(parameters, string_types):
            parameters = json.loads(parameters)

        execute_notebook(in_file, out_file, parameters=parameters, report_mode=hide_input, start_timeout=600)

        with open(out_file, 'r') as fp:
            output_text = fp.read()

    return output_text



import falcon
import uuid


class StorageEngine(object):
    '''Unified interface into storage classes of a backend'''

    ################################
    # These attributes must be set #
    users = property(lambda self: self._user_storage)
    notebooks = property(lambda self: self._notebook_storage)
    jobs = property(lambda self: self._job_storage)
    reports = property(lambda self: self._report_storage)
    ################################

    def __init__(self, user_storage, notebook_storage, job_storage, report_storage):
        # resources get with self.db
        self._user_storage = user_storage
        self._notebook_storage = notebook_storage
        self._job_storage = job_storage
        self._report_storage = report_storage

        # storage classes get with self.db
        self._user_storage.db = self
        self._notebook_storage.db = self
        self._job_storage.db = self
        self._report_storage.db = self

    def get_things(self, marker, limit):
        return [{'id': str(uuid.uuid4()), 'color': 'green'}]

    def add_thing(self, thing):
        thing['id'] = str(uuid.uuid4())
        return thing


class StorageError(Exception):

    @staticmethod
    def handle(ex, req, resp, params):
        description = ('Sorry, couldn\'t write your thing to the '
                       'database. It worked on my box.')

        raise falcon.HTTPError(falcon.HTTP_725,
                               'Database Error',
                               description)



import os
import os.path
import tempfile
from traitlets import HasTraits, Unicode, Bool
from ..scheduler import DummyScheduler, AirflowScheduler, LuigiScheduler, LocalScheduler


class SchedulerConfig(HasTraits):
    '''Base config for scheduler'''
    type = Unicode()


class DummySchedulerConfig(SchedulerConfig):
    '''Configuration for airflow'''
    type = 'dummy'
    clazz = DummyScheduler


class AirflowSchedulerConfig(SchedulerConfig):
    '''Configuration for airflow'''
    type = 'airflow'
    dagbag = Unicode(default_value=os.path.expanduser('~/airflow/dags'))
    config = Unicode(default_value=os.path.expanduser('~/airflow/airflow.cfg'))
    clazz = AirflowScheduler


class LuigiSchedulerConfig(SchedulerConfig):
    '''Configuration for luigi'''
    type = 'luigi'
    task_folder = Unicode(default_value=os.path.expanduser('~/luigi/tasks'))
    db_connection = Unicode(default_value=os.path.expanduser('sqlite://~/luigi/luigi-task-hist.db'))
    state_path = Unicode(default_value=os.path.expanduser('~/luigi/state.pickle'))

    # local or cron for scheduler
    local = Bool(default_value=True)

    crontab = Unicode(default_value=os.path.expanduser('~/luigi/cron.tab'))
    clazz = LuigiScheduler


class LocalSchedulerConfig(SchedulerConfig):
    '''Configuration for local scheduler'''
    type = 'local'
    working_directory = Unicode(default_value=os.path.join(tempfile.gettempdir(), 'paperboy'))
    clazz = LocalScheduler



import getpass
import os
import logging
import time
import threading
from datetime import datetime, timedelta
from concurrent.futures import ThreadPoolExecutor, as_completed
from queue import Queue, Empty
from ...worker import run


def interval_to_schedule(cronjob):
    '''this function will be called roughly every minute
    and should parse the cronjob's timing to see if the
    job should run in this minute'''
    now = datetime.now()

    # parse minutes first
    minute = str(cronjob.minute)
    hour = str(cronjob.hour)
    day_of_week = str(cronjob.dow)
    day_of_month = str(cronjob.dom)

    ret = (_parse(now.minute, minute),
           _parse(now.hour, hour),
           _parse(now.day, day_of_month),
           _parse_dow(now.strftime('%a'), day_of_week))
    if all(ret):
        return True
    return False


def _parse(now, field):
    # run every time?
    if str(field) == '*/1' or str(field) == '*':
        return True

    # else run every interval?
    if '-59' in field:
        splits = field.replace('-59', '').replace('-23', '').split('/')
        base, mod = int(splits[0]), int(splits[1])

        # check if now % mod == base
        if now % mod != base:
            return True
        else:
            return False

    # else run on specific minutes
    field = int(field)
    if field == now:
        return True
    else:
        # not the right minute
        return False


def _parse_dow(now, field):
    # run every time?
    if str(field) == '*' or str(field).upper() == now.upper():
        return True
    return False


def run_tasks(add_queue, delete_queue):
    executor = ThreadPoolExecutor(max_workers=os.cpu_count() - 1)
    tasks = {}
    futures = []
    while True:
        for tid, task in tasks.items():
            job, reports, job_dir, interval, cron = tasks[tid]
            if interval_to_schedule(cron):
                logging.critical('Submitting - %s' % str(job))
                futures.append(executor.submit(run, job, reports, job_dir))

        for future in as_completed(futures):
            logging.critical('Future finished')
            if future.exception():
                raise future.exception()
            else:
                logging.critical(future.result())

        now = datetime.now()
        end = datetime(year=now.year, month=now.month, day=now.day, hour=now.hour, minute=now.minute) + timedelta(minutes=1)
        while datetime.now() < end:
            # add to tasks
            while True:
                try:
                    item = add_queue.get_nowait()
                    tasks[item[0].id] = item
                except Empty:
                    break

                # delete tasks
            while True:
                try:
                    item = delete_queue.get_nowait()
                    tasks.pop(item, None)
                except Empty:
                    break
            time.sleep(1)


class LocalProcessScheduler(object):
    def __init__(self):
        self.tasks = {}
        self.add_queue = Queue()
        self.delete_queue = Queue()
        self.runner = None

    def schedule(self, job, reports, job_dir, interval):
        from crontab import CronTab, CronItem

        # must occur in gunicorn process
        if self.runner is None:
            self.runner = threading.Thread(target=run_tasks, args=(self.add_queue, self.delete_queue))
            self.runner.daemon = True
            self.runner.start()

        c = CronTab(user=getpass.getuser())
        cjob = CronItem.from_line(interval + ' echo test', cron=c)

        self.tasks[job.id] = (job, reports, job_dir, interval, cjob)
        logging.critical('Scheduling job: %s' % str(job.id))
        self.add_queue.put((job, reports, job_dir, interval, cjob))

    def unschedule(self, job_id):
        self.tasks[job_id] = None
        self.delete_queue.put(job_id)
